"""
Sandai Operator SDK

A Python SDK for building high-performance, asynchronous batch processing operators.
"""

from .processor import BatchProcessor
from .channel import FileChannel
from .types import (
    TaskInput, TaskOutput, TaskError, 
    InputArtifact, OutputArtifact, 
    ProcessingContext,
    OptionsType, ResultsType,
    EOFMarker,
)

__all__ = [
    "BatchProcessor",
    "FileChannel", 
    "TaskInput",
    "TaskOutput",
    "TaskError",
    "InputArtifact",
    "OutputArtifact",
    "ProcessingContext",
    "OptionsType",
    "ResultsType",
    "EOFMarker",
]